//
//  DBModel2.swift
//  AgendaAssistent
//
//  Created by André Hartman on 11/11/2023.
//  Copyright © 2023 André Hartman. All rights reserved.
//

import EventKit
import GRDB

var db = DBModel().getDB()

class dataModel {
    func getCalendarEvents(dates: PeriodStartEnd) -> [EKEvent] {
        var calendarEvents = [EKEvent]()
        let selectedCalendars: [EKCalendar] = eventStore.calendars(for: .event)
            .filter { $0.title.contains("Marieke") && !$0.title.contains("blokkeren") }
        let predicate = eventStore.predicateForEvents(withStart: dates.start, end: dates.end, calendars: selectedCalendars)
        calendarEvents = eventStore.events(matching: predicate)
            .filter { !$0.isAllDay }
        if calendarEvents.count == 0 {
            print("no events found")
            exit(0)
        }
        return calendarEvents
    }
}

class DBModel {
    public func getDB() -> DatabaseQueue {
        do {
            let fm = FileManager.default
            let path = try fm.url(for: .documentDirectory, in: .allDomainsMask, appropriateFor: nil, create: false)
            let databaseURL = path.appendingPathComponent("db.sqlite")
            var dbQueue = try DatabaseQueue(path: databaseURL.path)
            print("db stored at \(databaseURL.path)")
            var config = Configuration()

             config.prepareDatabase { db in
                 db.trace { print($0) }
             }

            dbQueue = try DatabaseQueue(path: databaseURL.path, configuration: config)
            return dbQueue
        } catch {
            fatalError("Unresolved error \(error)")
        }
    }

    public func makeDB() -> Int {
        var counter = 0
        do {
            // try db.erase()
            try db.write { db in
                try db.create(table: "patient", ifNotExists: true) { t in
                    t.autoIncrementedPrimaryKey("id")
                    t.column("patientName", .text)
                        .indexed()
                        .unique()
                        .notNull()
                    t.column("patientLatitude", .real)
                        .indexed()
                    t.column("patientLongitude", .real)
                }
            }
            try db.write { db in
                try db.create(table: "visit", ifNotExists: true) { t in
                    t.column("id", .text)
                        .primaryKey()
                        .indexed()
                    t.column("modifiedDate", .text)
                        .indexed()
                    t.column("visitAge", .integer)
                    t.column("visitCalendar", .text)
                    t.column("visitCreated", .text)
                    t.column("visitDate", .text)
                        .indexed()
                    t.column("visitFirst", .boolean)
                    t.column("visitRecurrent", .boolean)
                        .defaults(to: false)
                    t.column("visitWeekDate", .text)
                        .indexed()
                    t.belongsTo("patient", onDelete: .cascade)
                }
            }
            try db.read { db in
                counter = try PatientInfo.Patient.fetchCount(db)
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
        return counter
    }

    func updateFromPatients(patients: [PatientInfo], dates: PeriodStartEnd) {
        do {
            try db.inTransaction { db in
                for patient in patients {
                    var tempPatient = PatientInfo.Patient(
                        patientName: patient.patientName,
                        patientLatitude: patient.patientLatitude,
                        patientLongitude: patient.patientLongitude
                    )
                    tempPatient = try tempPatient.upsertAndFetch(
                        db, onConflict: ["patientName"],
                        doUpdate: { _ in
                            [Column("latitude").noOverwrite,
                             Column("longitude").noOverwrite]
                        }
                    )
                    for visit in patient.visits {
                        var tempVisit = visit
                        tempVisit.patientId = tempPatient.id!
                        try tempVisit.upsert(db)
                    }
                }
                return .commit
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
    }

    func deleteAllPatients() {
        do {
            try db.write { db in
                try PatientInfo.Patient.deleteAll(db)
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
    }

    func updateFromGeo(patients: [PatientInfo.Patient]) {
        do {
            try db.inTransaction { db in
                for p in patients {
                    if p.patientLatitude == 0.0 {
                        print("Faulty latitude (0)")
                    } else {
                        var patient = p
                        try patient.upsert(db)
                    }
                }
                return .commit
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
    }

    func getPatientInfo(dates: PeriodStartEnd) -> [PatientInfo] {
        var patientInfo = [PatientInfo]()
        do {
            patientInfo = try db.read { db in
                let filteredVisits = PatientInfo.Patient.visits
                    .filter(Column("visitDate") >= dates.start)
                    .filter(Column("visitDate") <= dates.end)
                return try PatientInfo.Patient
                    .including(all: filteredVisits)
                    .having(filteredVisits.isEmpty == false)
                    .asRequest(of: PatientInfo.self)
                    .fetchAll(db)
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
        return patientInfo
    }

    func getPatientForCSV(dates: PeriodStartEnd) -> [PatientInfo] {
        var patientInfo = [PatientInfo]()
        do {
            patientInfo = try db.read { db in
                let filteredVisits = PatientInfo.Patient.visits
                    .filter(Column("visitDate") >= dates.start)
                    .filter(Column("visitDate") <= dates.end)
                    .filter(["Marieke", "Marieke nieuwe"].contains(Column("visitCalendar")))
                return try PatientInfo.Patient
                    .including(all: filteredVisits)
                    .having(filteredVisits.isEmpty == false)
                    .asRequest(of: PatientInfo.self)
                    .fetchAll(db)
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
        return patientInfo
    }

    struct PatientTimelineInfo: Decodable, FetchableRecord {
        var patient: PatientInfo.Patient
        var minVisitVisitCreated: Date
        var maxVisitVisitDate: Date
    }

    func getPatientTimelineTest() {
        let temp = getPatientTimeline()
    }

    func getPatientTimeline() -> [PatientTimelineInfo] {
        var patientTimeline = [PatientTimelineInfo]()
        do {
            patientTimeline = try db.read { db in
                let filteredVisits = PatientInfo.Patient.visits
                    .filter(["Marieke", "Marieke nieuwe"].contains(Column("visitCalendar")))
                let request = PatientInfo.Patient
                    .select(
                        Column("id"),
                        Column("patientName")
                    )
                    .order(Column("patientName"))
                    //.having(filteredVisits.isEmpty == false)
                    .annotated(with:
                        filteredVisits.min(Column("visitCreated")),
                        filteredVisits.max(Column("visitDate")))
                return try PatientTimelineInfo.fetchAll(db, request)
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
        return patientTimeline
    }

    func getVisitMinMax() -> (Date?, Date?) {
        var minRow: Date?
        var maxRow: Date?

        do {
            try db.read { db in
                let request = PatientInfo.Visit.select(min(Column("visitDate")), max(Column("visitDate"))).filter(Column("visitRecurrent") == false)
                let row = try Row.fetchOne(db, request)!
                minRow = row[0]
                maxRow = row[1]
            }
        } catch {
            fatalError("Unresolved error \(error)")
        }
        return (minRow, maxRow)
    }

    func getNoGeoPatients() -> [PatientInfo.Patient] {
        var patients = [PatientInfo.Patient]()
        do {
            try db.read { db in
                patients = try PatientInfo.Patient
                    .filter(Column("patientLatitude") == nil)
                    .fetchAll(db)
            }
        } catch {
            fatalError("getNoGeoPatients: \(error)")
        }
        patients.shuffle()
        let count = min(patients.count, 199) - 1
        patients = Array(patients[...count])
        return patients
    }
}
